from typing import Callable, Union, List, Dict

import pandas as pd
import torch
import torchaudio
from torch.utils.data import Dataset
from torchaudio.compliance import kaldi
from tqdm import tqdm

from .features.padding import pad_dup
from .features.transform import NormalizeSample


class SERSliceDataset(Dataset):
    """Speech Emotion Recognition Dataset
    Storing all dataset in RAM
    """

    def __init__(
            self,
            csv_file: Union[str, pd.DataFrame],
            max_len: int,
            len_thresh: float = 0.5,
            pad_fn: Callable = pad_dup,
            sampling_rate: int = 16000,
            center_feats: bool = True,
            scale_feats: bool = True,
            emotions=None,
            transform=None):
        """
        Args:
            csv_file (string): Path to the csv file with annotations.
            max_len (int): Maximum length (in second) of audio file. Others will be truncated
            len_thresh (int): A threshold to cut audio sample that duration is shorter than provided length (in second)
            pad_fn (Callable): padding function from vistec_ser.data.features.padding
            sampling_rate (int): target sampling rate of audio. If this isn't match with file's sampling rate,
                result will be resampled into target sampling rate
            transform (callable, optional): Optional transform to be applied
                on a sample.
        """
        if emotions is None:
            self.emotions = ["neutral", "anger", "happiness", "sadness"]
        else:
            self.emotions = emotions
        self.n_classes = len(self.emotions)
        assert isinstance(max_len, int)
        assert isinstance(sampling_rate, int)
        self.sampling_rate = sampling_rate
        self.max_len = max_len * 100
        self.len_thresh = len_thresh
        self.pad_fn = pad_fn
        self.transform = transform
        self.normalize = NormalizeSample(center_feats, scale_feats)
        self.samples = self._load_csv(csv_file)

    def _chop_sample(self, sample: torch.Tensor) -> List[torch.Tensor]:
        """Chop spectrogram (or any 2-D Tensor) into smaller chunks based on `self.max_len`.
        Each chunk is normalized according to object's config (center_feats, scale_feats)s

        :param sample: Input tensor. Can be either spectrogram or fbank
        :return: a list of chopped samples
        """
        x, y = sample["feature"], sample["emotion"]
        _, time_dim = x.shape
        x_chopped = list()
        for i in range(time_dim):
            if i % self.max_len == 0 and i != 0:  # if reach self.max_len
                xi = x[:, i - self.max_len:i]
                assert xi.shape[-1] == self.max_len, xi.shape
                x_chopped.append(self.normalize({"feature": xi, "emotion": y}))
        if time_dim < self.max_len:  # if file length not reach self.max_len
            if self.pad_fn:
                xi = self.pad_fn(x, max_len=self.max_len)
                assert xi.shape[-1] == self.max_len
            else:
                xi = x
            x_chopped.append(self.normalize({"feature": xi, "emotion": y}))
        else:  # if file is longer than n_frame, pad remainder
            remainder = x[:, x.shape[-1] - x.shape[0] % self.max_len:]
            if not remainder.shape[-1] <= self.len_thresh:
                if self.pad_fn:
                    xi = self.pad_fn(remainder, max_len=self.max_len)
                else:
                    xi = x
                x_chopped.append(self.normalize({"feature": xi, "emotion": y}))
        return x_chopped

    def _load_feature(self, audio_path: str) -> torch.Tensor:
        audio, sample_rate = torchaudio.backend.sox_backend.load(audio_path)

        # initial preprocess
        # convert to mono, resample, truncate
        audio = torch.unsqueeze(audio.mean(dim=0), dim=0)  # convert to mono
        if sample_rate != self.sampling_rate:
            audio = kaldi.resample_waveform(audio, orig_freq=sample_rate, new_freq=self.sampling_rate)
        return audio

    def _load_csv(self, csv_file: Union[str, pd.DataFrame]) -> List[Dict[str, torch.Tensor]]:
        if isinstance(csv_file, str):
            csv = pd.read_csv(csv_file)
        else:
            csv = csv_file
        print("Extracting Features...")
        samples = []
        for i, (path, emotion) in tqdm(csv.iterrows(), total=len(csv)):
            if emotion.lower().strip() not in self.emotions:
                continue
            emotion = self.emotions.index(emotion.lower().strip())  # convert to label
            sample = self.transform({"feature": self._load_feature(path), "emotion": emotion})
            samples += self._chop_sample(sample)
        return samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        return self.samples[idx]


class SERSliceTestDataset(SERSliceDataset):
    def __init__(
            self,
            csv_file: Union[str, pd.DataFrame],
            max_len: int,
            sampling_rate: int = 16000,
            center_feats: bool = True,
            scale_feats: bool = False,
            emotions=None,
            transform=None):
        super().__init__(
            csv_file=csv_file,
            max_len=max_len,
            sampling_rate=sampling_rate,
            center_feats=center_feats,
            scale_feats=scale_feats,
            emotions=emotions,
            transform=transform)

    def _load_csv(self, csv_file):
        if isinstance(csv_file, str):
            csv = pd.read_csv(csv_file)
        else:
            csv = csv_file
        print("Extracting Features...")
        samples = []
        for i, (path, emotion) in tqdm(csv.iterrows(), total=len(csv)):
            if emotion.lower().strip() not in self.emotions:
                continue
            emotion = self.emotions.index(emotion.lower().strip())  # convert to label
            sample = self.transform({"feature": self._load_feature(path), "emotion": emotion})
            samples.append(self._chop_sample(sample))
        return samples


class SERInferenceDataset(SERSliceDataset):
    def __init__(
            self,
            csv_file: Union[str, pd.DataFrame],
            max_len: int,
            sampling_rate: int = 16000,
            center_feats: bool = True,
            scale_feats: bool = False,
            emotions=None,
            transform=None):
        super().__init__(
            csv_file=csv_file,
            max_len=max_len,
            sampling_rate=sampling_rate,
            center_feats=center_feats,
            scale_feats=scale_feats,
            emotions=emotions,
            transform=transform)

    def _load_csv(self, csv_file):
        if isinstance(csv_file, str):
            csv = pd.read_csv(csv_file)
        else:
            csv = csv_file
        print("Extracting Features...")
        samples = []
        for i, path in csv.iterrows():
            path = path["PATH"]
            # not a good practice to fill field emotion with PATH
            # but this is the easiest way to load and transform data without any
            # additional pipelines. Might consider fix in future
            sample = self.transform({"feature": self._load_feature(path), "emotion": path})
            samples.append(self._chop_sample(sample))
        return samples



from typing import Tuple
import os

import yaml


def load_yaml(config_path: str) -> dict:
    assert isinstance(config_path, str), "argument must be string"
    assert os.path.exists(config_path), f"config at `{config_path}` does not exists"
    assert config_path.split(".")[-1] in ["yaml", "yml"], f"Config file `{config_path}`"
    config = yaml.load(open(config_path, "r", encoding="utf-8"), Loader=yaml.SafeLoader)
    return config


def read_config(cfg: dict, test_fold: int = 0, include_zoom: bool = False) -> Tuple[dict, dict]:
    feature_config = cfg.get("feature", {"frame_shift": 10})
    thaiser_config = cfg.get("thaiser", {})
    module_params = {
        "test_fold": test_fold,
        "include_zoom": include_zoom,
        **thaiser_config, **feature_config}

    emotions = feature_config.get("emotions", ["neutral", "anger", "happiness", "sadness"])
    in_channel = feature_config.get("num_mel_bins", 40)
    sequence_length = feature_config.get("max_len", 3) * feature_config["frame_shift"] * 10
    model_config = cfg.get("cnn1dlstm", {})
    hparams = {"in_channel": in_channel, "sequence_length": sequence_length, "n_classes": len(emotions), **model_config}
    return hparams, module_params



import argparse
import warnings

from pytorch_lightning.loggers import TensorBoardLogger
import pytorch_lightning as pl

from vistec_ser.data.datasets.thaiser import ThaiSERDataModule
from vistec_ser.evaluation.evaluate import evaluate_slice_model
from vistec_ser.models.network import CNN1DLSTMSlice, CNN1DLSTMAttentionSlice
from vistec_ser.utils.utils import load_yaml, read_config
warnings.filterwarnings("ignore")


def run_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser(usage="Train AIS-SER Model")
    parser.add_argument("--config-path", "-cp",
                        default="examples/thaiser.yaml", type=str, help="Path to training config file")
    parser.add_argument("--include-zoom",
                        action="store_true", help="state whether to include Zoom Recording fold or not")
    parser.add_argument("--n-iter", "-n", default=25, type=int, help="Number of iteration")
    parser.add_argument("--attention", action="store_true", help="State whether to use attention LSTM or not")
    return parser.parse_args()


def run_fold(fold: int, config_path: str, include_zoom: bool, n_iter: int = 25, use_attn: bool = False):
    # load dataset & model
    config = load_yaml(config_path)
    hparams, module_params = read_config(config, test_fold=fold, include_zoom=include_zoom)
    thaiser_module = ThaiSERDataModule(**module_params)
    thaiser_module.set_fold(fold)  # set fold to evaluate

    # trainer
    trainer_config = config.get("trainer", {})
    callbacks = [
        pl.callbacks.ModelCheckpoint(
            save_top_k=5,
            monitor="val_acc",
            mode="min"
        ),
    ]

    # dataloader
    thaiser_module.setup()
    train_dataloader = thaiser_module.train_dataloader()
    val_dataloader = thaiser_module.val_dataloader()
    test_dataloader = thaiser_module.test_dataloader()
    if not include_zoom:
        zoom_dataloader = thaiser_module.zoom_dataloader()
    else:
        zoom_dataloader = None

    open(f"{thaiser_module.experiment_dir}/results.txt", "w").write("WeightedAccuracy,UnweightedAccuracy\n")
    open(f"{thaiser_module.experiment_dir}/confusion_matrix.txt", "w").write("")
    for i in range(n_iter):
        # reset model
        model = CNN1DLSTMAttentionSlice(hparams) if use_attn else CNN1DLSTMSlice(hparams)

        # trainer
        logger = TensorBoardLogger(
            save_dir=thaiser_module.experiment_dir,
            name="iterations",
            version=f"iteration_{i}")
        trainer = pl.Trainer(
            callbacks=callbacks,
            logger=logger,
            weights_save_path=thaiser_module.experiment_dir,
            **trainer_config)

        # train
        print(f"============= Running Experiment {i} =============")
        print("\n>>Training Model...\n")
        trainer.fit(model, train_dataloader=train_dataloader, val_dataloaders=val_dataloader)
        trainer.save_checkpoint(f"{thaiser_module.experiment_dir}/weights/final{i}.ckpt")

        # test
        print("\n>>Evaluating Model...\n")
        wa, ua, cm = evaluate_slice_model(model, test_dataloader, n_classes=thaiser_module.n_classes)
        cm = cm.numpy()
        template = f"Confusion Matrix:\n{cm}\nWeighted Accuracy: {wa*100:.2f}%\nUnweighted Accuracy: {ua*100:.2f}%"
        print(template)
        open(f"{thaiser_module.experiment_dir}/results.txt", "a").write(f"{wa*100:.2f},{ua*100:.2f}\n")
        open(f"{thaiser_module.experiment_dir}/confusion_matrix.txt", "a").write(f"Iteration {i}:\n{cm.numpy()}\n\n")

        # test zoom
        if not include_zoom:
            print("\n>>Evaluating Model (Zoom Test Set)...\n")
            wa, ua, cm = evaluate_slice_model(model, zoom_dataloader, n_classes=thaiser_module.n_classes)
            cm = cm.numpy()
            template = f"Confusion Matrix:\n{cm}\nWeighted Accuracy: {wa * 100:.2f}%\nUnweighted Accuracy: {ua * 100:.2f}%"
            print(template)
            open(f"{thaiser_module.experiment_dir}/results_zoom.txt", "a").write(f"{wa * 100:.2f},{ua * 100:.2f}\n")
            open(f"{thaiser_module.experiment_dir}/confusion_matrix_zoom.txt", "a").write(f"{cm.numpy()}")


def main(arguments):
    config_path = arguments.config_path
    include_zoom = arguments.include_zoom
    use_attn = arguments.attention

    config = load_yaml(config_path)
    thaiser_config = config.get("thaiser", {})
    thaiser_module = ThaiSERDataModule(test_fold=0, **thaiser_config)
    thaiser_module.prepare_data()

    for fold in thaiser_module.fold_config.keys():
        print(f"\n+-----------------------------------------+")
        print(f"| Experiment on fold {fold:02d}                   |")
        print(f"+-----------------------------------------+\n")
        run_fold(fold, config_path, use_attn=use_attn, include_zoom=include_zoom)


if __name__ == '__main__':
    args = run_parser()
    main(args)



import torch
import torchaudio
from torchaudio.compliance import kaldi


class Spectrogram(object):
    """Convert audio into spectrogram with STFT, using torchaudio.compliance.kaldi
    see more at https://pytorch.org/audio/stable/compliance.kaldi.html

    Args:
        frame_length (float, optional) – Frame length in milliseconds (Default: 25.0)
        frame_shift (float, optional) – Frame shift in milliseconds (Default: 10.0)
        preemphasis_coefficient (float, optional) – Coefficient for use in signal preemphasis (Default: 0.97)
        window_type (str, optional) – Type of window
            (‘hamming’|’hanning’|’povey’|’rectangular’|’blackman’) (Default: 'povey')
        sample_frequency (float, optional) – Waveform data sample frequency
            (must match the waveform file, if specified there) (Default: 16000.0)
        dither (float, optional) – Dithering constant (0.0 means no dither). If you turn this off,
            you should set the energy_floor option, e.g. to 1.0 or 0.1 (Default: 0.0)
    """
    def __init__(
            self,
            frame_length: float = 25.,
            frame_shift: float = 10.,
            preemphasis_coefficient: float = 0.97,
            window_type: str = "povey",
            sample_frequency: float = 16000.,
            dither: float = 0.):
        self.spectrogram_params = {
            "frame_length": frame_length,
            "frame_shift": frame_shift,
            "preemphasis_coefficient": preemphasis_coefficient,
            "window_type": window_type,
            "sample_frequency": sample_frequency,
            "dither": dither
        }

    def __call__(self, sample):
        audio, emotion = sample["feature"], sample["emotion"]
        spectrogram = kaldi.spectrogram(audio, **self.spectrogram_params)
        spectrogram = torch.transpose(spectrogram, 0, 1)
        return {"feature": spectrogram, "emotion": emotion}


class FilterBank(object):
    """Convert audio into fbank feature, using torchaudio.compliance.kaldi
        see more at https://pytorch.org/audio/stable/compliance.kaldi.html

        Args:
            frame_length (float, optional) – Frame length in milliseconds (Default: 25.0)
            frame_shift (float, optional) – Frame shift in milliseconds (Default: 10.0)
            num_mel_bins (int, optional) – Number of triangular mel-frequency bins (Default: 23)
            preemphasis_coefficient (float, optional) – Coefficient for use in signal preemphasis (Default: 0.97)
            window_type (str, optional) – Type of window
                (‘hamming’|’hanning’|’povey’|’rectangular’|’blackman’) (Default: 'povey')
            sample_frequency (float, optional) – Waveform data sample frequency
                (must match the waveform file, if specified there) (Default: 16000.0)
            dither (float, optional) – Dithering constant (0.0 means no dither). If you turn this off,
                you should set the energy_floor option, e.g. to 1.0 or 0.1 (Default: 0.0)
            high_freq (float, optional) – High cutoff frequency for mel bins
                (if <= 0, offset from Nyquist) (Default: 0.0)
            low_freq (float, optional) – Low cutoff frequency for mel bins (Default: 20.0)
        """

    def __init__(
            self,
            frame_length: float = 50.,
            frame_shift: float = 10.,
            num_mel_bins: int = 40,
            preemphasis_coefficient: float = 0.97,
            window_type: str = "hanning",
            sample_frequency: float = 16000.,
            dither: float = 0.,
            low_freq: float = None,
            high_freq: float = None,
            vtln_max: float = 1.0,
            vtln_min: float = 1.0):
        if high_freq is None:
            high_freq = sample_frequency // 2
        if low_freq is None:
            low_freq = 0.
        self.vtln_range = torch.arange(vtln_min, vtln_max, 0.01)
        self.spectrogram_params = {
            "frame_length": frame_length,
            "frame_shift": frame_shift,
            "num_mel_bins": num_mel_bins,
            "preemphasis_coefficient": preemphasis_coefficient,
            "window_type": window_type,
            "sample_frequency": sample_frequency,
            "dither": dither,
            "high_freq": high_freq,
            "low_freq": low_freq
        }

    def _sample_vtln_factor(self):
        if len(self.vtln_range) == 0:
            return 1.0
        else:
            idx = torch.randint(0, len(self.vtln_range), (1,))[0]
            return self.vtln_range[idx]

    def __call__(self, sample):
        audio, emotion = sample["feature"], sample["emotion"]
        alpha = self._sample_vtln_factor()
        fbank = kaldi.fbank(audio, vtln_warp=alpha, **self.spectrogram_params)
        fbank = torch.transpose(fbank, 0, 1)
        return {"feature": fbank, "emotion": emotion}


class NormalizeSample(object):
    def __init__(self,
                 center_feats: bool = True,
                 scale_feats: bool = False):
        self.center_feats = center_feats
        self.scale_feats = scale_feats

    def __call__(self, sample):
        feature, emotion = sample["feature"], sample["emotion"]
        if self.center_feats:
            # feature = feature - feature.mean(dim=0)
            feature = feature - torch.unsqueeze(feature.mean(dim=-1), dim=-1)
        if self.scale_feats:
            # feature = feature / torch.sqrt(feature.var(dim=0) + 1e-8)
            feature = feature / torch.sqrt(torch.unsqueeze(feature.var(dim=-1), dim=-1) + 1e-8)
        return {"feature": feature, "emotion": emotion}


class SpecAugment(object):
    """Time/Freq Masking

    Args:
        freq_mask_param (int): Parameter for masking frequency axis
        time_mask_param (int): Parameter for masking time axis
        n_time_mask (int): number of mask on time axis
        n_freq_mask (int): number of mask on frequency axis
    """
    def __init__(
            self,
            freq_mask_param: int,
            time_mask_param: int,
            n_time_mask: int,
            n_freq_mask: int):
        assert isinstance(freq_mask_param, int)
        assert isinstance(time_mask_param, int)
        assert isinstance(n_freq_mask, int)
        assert isinstance(n_time_mask, int)

        self.freq_mask_param = freq_mask_param
        self.time_mask_param = time_mask_param
        self.n_time_mask = n_time_mask
        self.n_freq_mask = n_freq_mask

        self.freq_masking = torchaudio.transforms.FrequencyMasking(freq_mask_param=freq_mask_param)
        self.time_masking = torchaudio.transforms.TimeMasking(time_mask_param=time_mask_param)

    def __call__(self, sample):
        spec, emotion = sample["feature"], sample["emotion"]
        for _ in range(self.n_freq_mask):
            spec = self.freq_masking(spec)
        for _ in range(self.n_time_mask):
            spec = self.time_masking(spec)
        return {"feature": spec, "emotion": emotion}



import argparse
import warnings

from pytorch_lightning.loggers import TensorBoardLogger
import pytorch_lightning as pl

from vistec_ser.data.datasets.thaiser import ThaiSERDataModule
from vistec_ser.models.network import CNN1DLSTMSlice
from vistec_ser.evaluation.evaluate import evaluate_slice_model
from vistec_ser.utils.utils import load_yaml, read_config
warnings.filterwarnings("ignore")


def run_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser(usage="Train AIS-SER Model")
    parser.add_argument("--fold", default=0, type=int, help="Fold number used as val/test set")
    parser.add_argument("--include-zoom",
                        action="store_true", help="state whether to include Zoom Recording fold or not")
    parser.add_argument("--config-path", "-cp",
                        default="examples/thaiser.yaml", type=str, help="Path to training config file")
    return parser.parse_args()


def main(arguments):
    # load dataset & model
    config_path = arguments.config_path
    test_fold = arguments.fold
    include_zoom = arguments.include_zoom
    config = load_yaml(config_path)

    hparams, module_params = read_config(config, test_fold=test_fold, include_zoom=include_zoom)
    trainer_config = config.get("trainer", {})

    model = CNN1DLSTMSlice(hparams)
    thaiser_module = ThaiSERDataModule(**module_params)
    thaiser_module.prepare_data()

    # dataloader
    thaiser_module.setup()
    train_dataloader = thaiser_module.train_dataloader()
    val_dataloader = thaiser_module.val_dataloader()
    test_dataloader = thaiser_module.test_dataloader()
    if not include_zoom:
        zoom_dataloader = thaiser_module.zoom_dataloader()
    else:
        zoom_dataloader = None

    open(f"{thaiser_module.experiment_dir}/results.txt", "w").write("WeightedAccuracy,UnweightedAccuracy\n")
    open(f"{thaiser_module.experiment_dir}/confusion_matrix.txt", "w").write("")
    if not include_zoom:
        open(f"{thaiser_module.experiment_dir}/results_zoom.txt", "w").write("")

    # trainer
    callbacks = [
        pl.callbacks.ModelCheckpoint(
            save_top_k=5,
            monitor="val_acc",
            mode="min")
    ]
    logger = TensorBoardLogger(save_dir=thaiser_module.experiment_dir)
    trainer = pl.Trainer(
        callbacks=callbacks,
        logger=logger,
        weights_save_path=thaiser_module.experiment_dir,
        **trainer_config)

    # train
    print("\n>>Training Model...\n")
    trainer.fit(model, train_dataloader=train_dataloader, val_dataloaders=val_dataloader)
    trainer.save_checkpoint(f"{thaiser_module.experiment_dir}/weights/final.ckpt")

    # test
    print("\n>>Evaluating Model...\n")
    wa, ua, cm = evaluate_slice_model(model, test_dataloader, n_classes=thaiser_module.n_classes)
    template = f"Confusion Matrix:\n{cm.numpy()}\nWeighted Accuracy: {wa*100:.2f}%\nUnweighted Accuracy: {ua*100:.2f}%"
    print(template)
    open(f"{thaiser_module.experiment_dir}/results.txt", "a").write(f"{wa * 100:.2f},{ua * 100:.2f}\n")
    open(f"{thaiser_module.experiment_dir}/confusion_matrix.txt", "a").write(f"{cm.numpy()}")

    # test zoom
    if not include_zoom:
        print("\n>>Evaluating Model (Zoom Test Set)...\n")
        wa, ua, cm = evaluate_slice_model(model, zoom_dataloader, n_classes=thaiser_module.n_classes)
        cm = cm.numpy()
        template = f"Confusion Matrix:\n{cm}\nWeighted Accuracy: {wa * 100:.2f}%\nUnweighted Accuracy: {ua * 100:.2f}%"
        print(template)
        open(f"{thaiser_module.experiment_dir}/results_zoom.txt", "a").write(f"{wa * 100:.2f},{ua * 100:.2f}\n")
        open(f"{thaiser_module.experiment_dir}/confusion_matrix_zoom.txt", "a").write(f"{cm.numpy()}")


if __name__ == '__main__':
    args = run_parser()
    main(args)



import torch
import pytorch_lightning as pl
import torch.nn.functional as F
from pytorch_lightning.metrics import functional as FM
from torch.optim import Adam


class BaseModel(pl.LightningModule):
    def __init__(self, hparams, **kwargs):
        super().__init__(**kwargs)
        self.hyperparams = hparams
        
    def forward(self, *args, **kwargs):
        raise NotImplementedError()

    def training_step(self, batch, batch_idx):
        x, y = batch["feature"], batch["emotion"]
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)
        preds = torch.argmax(y_hat, dim=-1)
        acc = FM.accuracy(preds, y)
        metrics = {"train_loss": loss, "train_acc": acc}
        self.log_dict(metrics, prog_bar=True, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch["feature"], batch["emotion"]
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)
        preds = torch.argmax(y_hat, dim=-1)
        acc = FM.accuracy(preds, y)
        metrics = {"val_acc": acc, "val_loss": loss}
        self.log_dict(metrics, prog_bar=True, logger=True)
        return metrics

    def test_step(self, batch, batch_idx):
        metrics = self.validation_step(batch, batch_idx)
        metrics = {'test_acc': metrics['val_acc'], 'test_loss': metrics['val_loss']}
        self.log_dict(metrics, prog_bar=True, logger=True)

    def configure_optimizers(self):
        opt = Adam(self.parameters(), lr=self.hyperparams.learning_rate)
        return opt


class BaseSliceModel(BaseModel):

    def __init__(self, hparams, **kwargs):
        super().__init__(hparams, **kwargs)

    def forward(self, *args, **kwargs):
        raise NotImplementedError()

    def validation_step(self, batch, batch_idx):
        emotion = batch[0]["emotion"]
        final_logits = torch.stack([self(chunk["feature"])[0] for chunk in batch]).mean(dim=0, keepdim=True)
        loss = F.cross_entropy(final_logits, emotion)
        acc = FM.accuracy(final_logits.argmax(dim=-1), emotion)
        metrics = {"val_acc": acc, "val_loss": loss}
        self.log_dict(metrics)
        return metrics



from collections import OrderedDict

import torch.nn as nn

from .base_model import BaseSliceModel, BaseModel
from .layers.rnn import AttentionLSTM


class CNN1DLSTM(BaseModel):
    def __init__(self, hparams):
        super().__init__(hparams)
        in_channel = hparams.get("in_channel", 40)
        sequence_length = hparams.get("sequence_length", 300)
        n_channels = hparams.get("n_channels", [64, 64, 128, 128])
        kernel_size = hparams.get("kernel_size", [5, 3, 3, 3])
        pool_size = hparams.get("pool_size", [4, 2, 2, 2])
        lstm_unit = hparams.get("lstm_unit", 128)
        n_classes = hparams.get("n_classes", 4)

        self.in_channel = in_channel
        self.sequence_length = sequence_length
        assert len(n_channels) == len(kernel_size) == len(pool_size), "Size of `n_channels`, `kernel_size`, and " \
                                                                      "`pool_size` must equal "

        # configure cnn parameters
        in_channels = [in_channel] + n_channels[:-1]
        out_channels = n_channels
        seq_lens = []
        for p in pool_size:
            seq_lens.append(sequence_length)
            sequence_length = sequence_length // p
        assert len(in_channels) == len(out_channels) == len(seq_lens)

        self.cnn_layers = nn.Sequential(OrderedDict([
            (f"conv{i}", nn.Sequential(
                nn.Conv1d(in_channels=ic, out_channels=oc, kernel_size=k, padding=(k - 1) // 2),
                nn.LeakyReLU(),
                nn.LayerNorm(normalized_shape=[oc, seq]),
                nn.MaxPool1d(kernel_size=p)
            )) for i, (ic, oc, k, seq, p) in enumerate(zip(in_channels, out_channels, kernel_size, seq_lens, pool_size))
        ]))
        self.lstm = nn.LSTM(input_size=out_channels[-1], hidden_size=lstm_unit, bidirectional=True, batch_first=True)
        self.logits = nn.Linear(lstm_unit * 2, n_classes)

    def forward(self, x):
        for cnn in self.cnn_layers:
            x = cnn(x)
        x = x.transpose(1, 2)
        _, (x, _) = self.lstm(x)
        x = x.transpose(0, 1)
        x = x.reshape(x.shape[0], x.shape[1] * x.shape[2])
        x = self.logits(x)
        return x


class CNN1DLSTMSlice(BaseSliceModel):
    def __init__(self, hparams):
        super().__init__(hparams)
        in_channel = hparams.get("in_channel", 40)
        sequence_length = hparams.get("sequence_length", 300)
        n_channels = hparams.get("n_channels", [64, 64, 128, 128])
        kernel_size = hparams.get("kernel_size", [5, 3, 3, 3])
        pool_size = hparams.get("pool_size", [4, 2, 2, 2])
        lstm_unit = hparams.get("lstm_unit", 128)
        n_classes = hparams.get("n_classes", 4)

        self.in_channel = in_channel
        self.sequence_length = sequence_length
        assert len(n_channels) == len(kernel_size) == len(pool_size), "Size of `n_channels`, `kernel_size`, and " \
                                                                      "`pool_size` must equal "

        # configure cnn parameters
        in_channels = [in_channel] + n_channels[:-1]
        out_channels = n_channels
        seq_lens = []
        for p in pool_size:
            seq_lens.append(sequence_length)
            sequence_length = sequence_length // p
        assert len(in_channels) == len(out_channels) == len(seq_lens)

        self.cnn_layers = nn.Sequential(OrderedDict([
            (f"conv{i}", nn.Sequential(
                nn.Conv1d(in_channels=ic, out_channels=oc, kernel_size=k, padding=(k - 1) // 2),
                nn.LeakyReLU(),
                nn.LayerNorm(normalized_shape=[oc, seq]),
                nn.MaxPool1d(kernel_size=p)
            )) for i, (ic, oc, k, seq, p) in enumerate(zip(in_channels, out_channels, kernel_size, seq_lens, pool_size))
        ]))
        self.lstm = nn.LSTM(input_size=out_channels[-1], hidden_size=lstm_unit, bidirectional=True, batch_first=True)
        self.logits = nn.Linear(lstm_unit * 2, n_classes)

    def forward(self, x):
        for cnn in self.cnn_layers:
            x = cnn(x)
        x = x.transpose(1, 2)
        _, (x, _) = self.lstm(x)
        x = x.transpose(0, 1)
        x = x.reshape(x.shape[0], x.shape[1] * x.shape[2])
        x = self.logits(x)
        return x


class CNN1DLSTMAttentionSlice(BaseSliceModel):
    def __init__(self, hparams):
        super().__init__(hparams)
        in_channel = hparams.get("in_channel", 40)
        sequence_length = hparams.get("sequence_length", 300)
        n_channels = hparams.get("n_channels", [64, 64, 128, 128])
        kernel_size = hparams.get("kernel_size", [5, 3, 3, 3])
        pool_size = hparams.get("pool_size", [4, 2, 2, 2])
        lstm_unit = hparams.get("lstm_unit", 128)
        n_classes = hparams.get("n_classes", 4)

        self.in_channel = in_channel
        self.sequence_length = sequence_length
        assert len(n_channels) == len(kernel_size) == len(pool_size), "Size of `n_channels`, `kernel_size`, and " \
                                                                      "`pool_size` must equal "

        # configure cnn parameters
        in_channels = [in_channel] + n_channels[:-1]
        out_channels = n_channels
        seq_lens = []
        for p in pool_size:
            seq_lens.append(sequence_length)
            sequence_length = sequence_length // p
        assert len(in_channels) == len(out_channels) == len(seq_lens)

        self.cnn_layers = nn.Sequential(OrderedDict([
            (f"conv{i}", nn.Sequential(
                nn.Conv1d(in_channels=ic, out_channels=oc, kernel_size=k, padding=(k - 1) // 2),
                nn.LeakyReLU(),
                nn.LayerNorm(normalized_shape=[oc, seq]),
                nn.MaxPool1d(kernel_size=p)
            )) for i, (ic, oc, k, seq, p) in enumerate(zip(in_channels, out_channels, kernel_size, seq_lens, pool_size))
        ]))
        self.lstm = AttentionLSTM(
            input_dim=out_channels[-1], hidden_dim=lstm_unit, bidirectional=True, output_dim=n_classes)

    def forward(self, x):
        for cnn in self.cnn_layers:
            x = cnn(x)
        x = x.transpose(1, 2)
        x = self.lstm(x)
        return x



from typing import List
import os

from fastapi import FastAPI, File, UploadFile
import aiofiles

from vistec_ser.inference.inference import infer_sample, setup_server

# setup model
config_path = "thaiser.yaml"

app = FastAPI()
model, thaiser_module, temp_dir = setup_server(config_path)


def clear_audio(audio_paths: List[str]) -> None:
    for f in audio_paths:
        os.remove(f)


@app.get("/healthcheck")
async def healthcheck():
    return {"status": "healthy"}


@app.post("/predict")
async def predict(audios: List[UploadFile] = File(...)):
    """
    Predict audio POST from front-end server using `form-data` files

    NOTE: note that this might bug if > 1 requests are sent with the same file name
    """
    # save files
    audio_paths = []
    for audio in audios:
        print(audio.filename)
        save_name = f"{temp_dir}/{audio.filename}"
        async with aiofiles.open(save_name, "wb") as f:
            content = await audio.read()
            await f.write(content)
        audio_paths.append(save_name)
        assert os.path.exists(save_name)

    # extract features
    inference_loader = thaiser_module.extract_feature(audio_paths)
    inference_results = [infer_sample(model, sample, emotions=thaiser_module.emotions)
                         for sample in inference_loader]

    clear_audio(audio_paths)
    return inference_results



from typing import Tuple

from torch.utils.data import DataLoader
import pytorch_lightning.metrics.functional as FM
import torch

from ..models.base_model import BaseSliceModel


def evaluate_slice_model(
        model: BaseSliceModel,
        test_dataloader: DataLoader,
        n_classes: int = 4) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    y_true, y_pred = [], []
    for batch_idx, batch in enumerate(test_dataloader):
        emotion = batch[0]["emotion"]
        final_logits = []
        for chunk in batch:
            logits = model(chunk["feature"])  # dim=(1, 4)
            final_logits.append(logits[0])
        prediction = torch.stack(final_logits).mean(dim=0).argmax(dim=-1, keepdim=True)
        y_true.append(emotion)
        y_pred.append(prediction)
    y_true = torch.stack(y_true).squeeze(-1)
    y_pred = torch.stack(y_pred).squeeze(-1)
    wa = FM.accuracy(y_pred, y_true)
    cm = FM.confusion_matrix(y_pred, y_true, normalize='true', num_classes=n_classes)
    ua = torch.diag(cm).mean()
    return wa, ua, cm



from typing import Callable, List

import torch


def pad_dup(x: torch.Tensor, max_len: int) -> torch.Tensor:
    """Pad an Arguments feature upto specified length.
    The Arguments is repeated until max_len is reached.
    """
    time_dim = x.shape[-1]
    tmp = x.clone()
    num_repeat = int(max_len / time_dim)
    remainder = max_len - num_repeat * time_dim
    x_rem = x[:, :remainder]
    for _ in range(num_repeat - 1):
        x = torch.cat([x, tmp], dim=-1)
    x_pad = torch.cat([x, x_rem], dim=-1)
    return x_pad


def pad_zero(x: torch.Tensor, max_len: int) -> torch.Tensor:
    """Pad Arguments feature up to specified length.
    The padded values are zero.
    Arguments
    """
    zeros = torch.zeros([x.shape[0], max_len - x.shape[1]])
    x_pad = torch.cat([x, zeros], dim=-1)
    return x_pad


def pad_X(X: List[torch.Tensor], pad_fn: Callable, max_len: int = None) -> torch.Tensor:
    """Pad a pack of array to a specified max_len.
    If max_len is not specified, longest preprocessing will
    be use as a max length. This function is used to
    pad the fbank features to its max_len, not chopped.
    Arguments
    """
    if not max_len:
        max_len = max([x.shape[-1] for x in X])
    out = [pad_fn(x, max_len) for x in X]
    return torch.stack(out)



import warnings
from glob import glob
from typing import Dict, List, Union
import json
import os
import zipfile
from urllib.error import HTTPError

from torch.utils.data import DataLoader
from torchvision.transforms.transforms import Compose
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import wget

from ..ser_slice_dataset import SERSliceDataset, SERSliceTestDataset, SERInferenceDataset
from ..features.transform import FilterBank

emo2idx = {emo: i for i, emo in enumerate(['Neutral', 'Angry', 'Happy', 'Sad', 'Frustrated'])}
idx2emo = {v: k for k, v in emo2idx.items()}
correctemo = {
    'Neutral': 'Neutral',
    'Angry': 'Anger',
    'Happy': 'Happiness',
    'Sad': 'Sadness',
    'Frustrated': 'Frustration'
}


class ThaiSERDataModule(pl.LightningDataModule):
    """
    THAI SER data module

    This class instantiate an object that allows an integration with
    pytorch lightning's Trainer module

    Attributes
    ---------
    test_fold: int
        fold number of dataset used as test set. This dataset is separated
        into 10 folds. See self.fold_config() for more info.
    agreement_threshold: float
        agreement threshold of the dataset. The dataset consists of multiple
        annotators. Thus, some data might have different annotation consensus
        among multiple annotators. `agreement_threshold` param is a threshold
        that will filter out label that multiple annotators did not agree upon
        selected ratio (default as 0.71). The default value was calculated from
        inter-rater score. We recommend not to change this value.
    sampling_rate: int
        target sampling rate of data. Original sound sampling rate was 44100 but
        default as 16000
    num_mel_bins: int
        number of mel bins used for extracting filterbank features
    frame_length: int
        frame length in ms of window used  for STFT
    frame_shift: int
        frame shift in ms of window used for STFT
    center_feats: bool
        specify whether to center filterbank features or not (shift to zero mean)
    scale_feats: bool
        specify whether to scale filterbank features or not (scale to unit variance)
    mic_type: str
        microphone type used for training dataset. Two choices avilable: `con` and `clip`,
        where `con` denotes condensor mic and `clip` denotes clip mic
    download_dir: str
        dataset download directory
    experiment_dir: str
        experiment directory
    include_zoom: bool
        specify whether to include zoom experiment in dataset fold or not
    max_len: int
        max context length of features
    batch_size: int
        batch size of training data
    emotions: List[str]
        list of allowed emotion. There are 5 emotions available
        [neutral, anger, happiness, sadness, frustration] but default
        as four emotions: neutral, anger, happiness, sadness
    """
    def __init__(
            self,
            test_fold: int,
            agreement_threshold: float = 0.71,
            sampling_rate: int = 16000,
            num_mel_bins: int = 40,
            frame_length: int = 50,  # in ms
            frame_shift: int = 10,  # in ms
            center_feats: bool = True,
            scale_feats: bool = True,
            mic_type: str = 'con',
            download_dir: str = None,
            experiment_dir: str = None,
            include_zoom: bool = True,
            max_len: int = 3,
            batch_size: int = 64,
            emotions: List[str] = None,
            num_workers: int = 0,
            *args,
            **kwargs):
        """
        Construct necessary attributes for THAI SER Data Module

        Parameters
        ----------
            test_fold: int
                fold number of dataset used as test set. This dataset is separated
                into 10 folds. See self.fold_config() for more info.
            agreement_threshold: float
                agreement threshold of the dataset. The dataset consists of multiple
                annotators. Thus, some data might have different annotation consensus
                among multiple annotators. `agreement_threshold` param is a threshold
                that will filter out label that multiple annotators did not agree upon
                selected ratio (default as 0.71). The default value was calculated from
                inter-rater score. We recommend not to change this value.
            sampling_rate: int
                target sampling rate of data. Original sound sampling rate was 44100 but
                default as 16000
            num_mel_bins: int
                number of mel bins used for extracting filterbank features
            frame_length: int
                frame length in ms of window used  for STFT
            frame_shift: int
                frame shift in ms of window used for STFT
            center_feats: bool
                specify whether to center filterbank features or not (shift to zero mean)
            scale_feats: bool
                specify whether to scale filterbank features or not (scale to unit variance)
            mic_type: str
                microphone type used for training dataset. Two choices avilable: `con` and `clip`,
                where `con` denotes condensor mic and `clip` denotes clip mic
            download_dir: str
                dataset download directory
            experiment_dir: str
                experiment directory
            include_zoom: bool
                specify whether to include zoom experiment in dataset fold or not
            max_len: int
                max context length of features
            batch_size: int
                batch size of training data
            emotions: List[str]
                list of allowed emotion. There are 5 emotions available
                [neutral, anger, happiness, sadness, frustration] but default
                as four emotions: neutral, anger, happiness, sadness
        """
        super().__init__(*args, **kwargs)
        if emotions is None:
            emotions = ["neutral", "anger", "happiness", "sadness"]
        # loading dataset config
        self.agreement_threshold = agreement_threshold
        self.mic_type = mic_type
        self.test_fold = test_fold

        # dataset config
        self.max_len = max_len
        self.batch_size = batch_size
        self.sampling_rate = sampling_rate
        self.frame_length = frame_length
        self.frame_shift = frame_shift
        self.sec_to_frame = 10 * self.frame_shift
        self.num_mel_bins = num_mel_bins
        self.num_workers = num_workers

        # normalizing sample
        self.center_feats = center_feats
        self.scale_feats = scale_feats

        # config n_classes, avail emotion
        self.include_zoom = include_zoom
        self.emotions = emotions
        self.n_classes = len(self.emotions)

        # config download dir
        if download_dir is None:
            self.download_root = f"{os.path.expanduser('~')}/vistec-ser_tmpfiles/vistec"
        else:
            self.download_root = f"{download_dir}/vistec-ser_tmpfiles/vistec"
        if not os.path.exists(self.download_root):
            os.makedirs(self.download_root)

        # config experiment dir
        if experiment_dir is None:
            self.experiment_root = f"{os.path.expanduser('~')}/vistec-ser_tmpfiles/exp_vistec"
        else:
            self.experiment_root = f"{experiment_dir}"
        self.experiment_dir = f"{self.experiment_root}/fold{self.test_fold}"
        if not os.path.exists(self.experiment_dir):
            os.makedirs(self.experiment_dir)

        # define download URL
        version = 1
        release_url = f"https://github.com/vistec-AI/dataset-releases/releases/download/v{version}"
        self.github_url = {
            "studio1-10": release_url+f"/studio1-10.zip",
            "studio11-20": release_url+f"/studio11-20.zip",
            "studio21-30": release_url+f"/studio21-30.zip",
            "studio31-40": release_url+f"/studio31-40.zip",
            "studio41-50": release_url+f"/studio41-50.zip",
            "studio51-60": release_url+f"/studio51-60.zip",
            "studio61-70": release_url+f"/studio61-70.zip",
            "studio71-80": release_url+f"/studio71-80.zip",
            "zoom1-10": release_url+f"/zoom1-10.zip",
            "zoom11-20": release_url+f"/zoom11-20.zip",
        }
        self.labels_url = release_url+f"/emotion_label.json"

        # define fold split
        self.fold_config = {
            0: [f"studio{s:03d}" for s in range(1, 11)],
            1: [f"studio{s:03d}" for s in range(11, 21)],
            2: [f"studio{s:03d}" for s in range(21, 31)],
            3: [f"studio{s:03d}" for s in range(31, 41)],
            4: [f"studio{s:03d}" for s in range(41, 51)],
            5: [f"studio{s:03d}" for s in range(51, 61)],
            6: [f"studio{s:03d}" for s in range(61, 71)],
            7: [f"studio{s:03d}" for s in range(71, 81)],
            8: [f"zoom{s:03d}" for s in range(1, 11)],
            9: [f"zoom{s:03d}" for s in range(11, 21)]
        }
        assert self.test_fold in self.fold_config.keys()
        self.studio_list = []
        for studios in self.fold_config.values():
            for s in studios:
                self.studio_list.append(s)

        self.train = None
        self.val = None
        self.test = None
        self.zoom = None

    def set_fold(self, fold):
        self.test_fold = fold
        self.experiment_dir = f"{self.experiment_root}/fold{self.test_fold}"
        if not os.path.exists(self.experiment_dir):
            os.makedirs(self.experiment_dir)

    def prepare_data(self):
        """Run once as a preparation: download dataset, generate csv labels"""
        self._download()
        self._prepare_labels()

    def setup(self, *args, **kwargs):
        train_folds = self.fold_config.keys() if self.include_zoom else list(self.fold_config.keys())[:-2]
        self.train = pd.concat([pd.read_csv(f"{self.download_root}/fold{i}.csv")
                                for i in train_folds if i != self.test_fold])
        test_split = pd.read_csv(f"{self.download_root}/fold{self.test_fold}.csv")
        test_studio = self.fold_config[self.test_fold]
        val_studio = test_studio[:len(test_studio) // 2]
        test_studio = test_studio[len(test_studio) // 2:]
        self.val = test_split[test_split["PATH"].apply(lambda x: x.split("/")[-3]).isin(val_studio)]
        self.test = test_split[test_split["PATH"].apply(lambda x: x.split("/")[-3]).isin(test_studio)]
        if not self.include_zoom:
            self.zoom = pd.concat([pd.read_csv(f"{self.download_root}/fold{i}.csv")
                                   for i in list(self.fold_config.keys())[-2:] if i != self.test_fold])

    def train_dataloader(self) -> DataLoader:
        transform = Compose([FilterBank(
            frame_length=self.frame_length,
            frame_shift=self.frame_shift,
            num_mel_bins=self.num_mel_bins)])
        train_vistec = SERSliceDataset(
            csv_file=self.train,
            sampling_rate=self.sampling_rate,
            max_len=self.max_len,
            center_feats=self.center_feats,
            scale_feats=self.scale_feats,
            emotions=self.emotions,
            transform=transform)
        return DataLoader(train_vistec, batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True)

    def val_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        transform = Compose([FilterBank(
                    frame_length=self.frame_length,
                    frame_shift=self.frame_shift,
                    num_mel_bins=self.num_mel_bins)])
        val_vistec = SERSliceTestDataset(
            csv_file=self.val,
            sampling_rate=self.sampling_rate,
            max_len=self.max_len,
            center_feats=self.center_feats,
            scale_feats=self.scale_feats,
            emotions=self.emotions,
            transform=transform)
        return DataLoader(val_vistec, batch_size=1, num_workers=self.num_workers)

    def test_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        transform = Compose([FilterBank(
                    frame_length=self.frame_length,
                    frame_shift=self.frame_shift,
                    num_mel_bins=self.num_mel_bins)])
        test_vistec = SERSliceTestDataset(
            csv_file=self.test,
            sampling_rate=self.sampling_rate,
            max_len=self.max_len,
            center_feats=self.center_feats,
            scale_feats=self.scale_feats,
            emotions=self.emotions,
            transform=transform
        )
        return DataLoader(test_vistec, batch_size=1, num_workers=self.num_workers)

    def zoom_dataloader(self) -> DataLoader:
        transform = Compose([FilterBank(
            frame_length=self.frame_length,
            frame_shift=self.frame_shift,
            num_mel_bins=self.num_mel_bins)])
        zoom_vistec = SERSliceTestDataset(
            csv_file=self.zoom,
            sampling_rate=self.sampling_rate,
            max_len=self.max_len,
            center_feats=self.center_feats,
            scale_feats=self.scale_feats,
            emotions=self.emotions,
            transform=transform
        )
        return DataLoader(zoom_vistec, batch_size=1, num_workers=self.num_workers)

    def extract_feature(self, audio_path: Union[str, List[str]]):
        # make audio_path List[str]
        if isinstance(audio_path, str):
            audio_path = [audio_path]
        audio_df = pd.DataFrame([[a] for a in audio_path], columns=["PATH"])
        transform = Compose([
            FilterBank(
                frame_length=self.frame_length,
                frame_shift=self.frame_shift,
                num_mel_bins=self.num_mel_bins
            )
        ])
        feature_dataset = SERInferenceDataset(
            csv_file=audio_df,
            sampling_rate=self.sampling_rate,
            max_len=self.max_len,
            center_feats=self.center_feats,
            scale_feats=self.scale_feats,
            transform=transform
        )
        return DataLoader(feature_dataset, batch_size=1, num_workers=self.num_workers)

    def _get_audio_path(self, audio_name: str) -> Union[str, None]:
        if not isinstance(audio_name, str):
            raise TypeError(f"audio name must be string but got {type(audio_name)}")
        studio_type = audio_name[0]
        studio_num = audio_name.split('_')[0][1:]
        if studio_type == "s":
            directory = f"studio{studio_num}"
        elif studio_type == "z":
            directory = f"zoom{studio_num}"
        else:
            raise NameError(f"Error reading file name {audio_name}")
        audio_path = f"{self.download_root}/{directory}/con/{audio_name}".replace(".wav", ".flac")
        if studio_type == "s":
            audio_path = audio_path.replace("con", self.mic_type)
        elif studio_type == "z":
            audio_path = audio_path.replace("con", "mic")
        else:
            raise NameError(f"Error reading file name {audio_name}")
        if not os.path.exists(audio_path):
            warnings.warn(f"{audio_path} not found, skipping...")
            return None
        return audio_path

    def _prepare_labels(self):
        # format
        if not os.path.exists(f"{self.download_root}/labels.csv"):
            print("\n+-----------------------------------+")
            print("| Formatting labels...              |")
            print("+-----------------------------------+")

            json_path = f"{self.download_root}/labels.json"
            if not os.path.exists(json_path):
                raise FileNotFoundError(f"labels.json not found at {self.download_root}")

            print(f">formatting {json_path} ...")
            data = read_json(json_path)
            # Filter studio that doesn't appear in download_dir
            avail_studio = []
            for std in sorted(glob(f"{self.download_root}/*/")):
                std = std[:-1].split("/")[-1]
                std = std[0] + std[-3:]
                avail_studio.append(std)
            data = {k: v for k, v in data.items() if k.split("_")[0] in avail_studio}
            agreements = get_agreements(data)
            labels = pd.DataFrame([
                (f"{self._get_audio_path(k)}", correctemo[idx2emo[v]])
                for k, v in {k: convert_to_hardlabel(v, thresh=self.agreement_threshold)
                             for k, v in agreements.items()}.items()
                if v != -1 and self._get_audio_path(k) is not None
            ], columns=['PATH', 'EMOTION'])

            labels.to_csv(f"{self.download_root}/labels.csv", index=False)
        else:
            labels = pd.read_csv(f"{self.download_root}/labels.csv")
        if all(os.path.exists(f"{self.download_root}/fold{fold}.csv") for fold in self.fold_config.keys()):
            pass
        else:
            print("\n+-----------------------------------+")
            print("| Separating folds...               |")
            print("+-----------------------------------+")
            studio = labels["PATH"].apply(lambda x: x.split("/")[-3])
            for n_fold, studio_num in self.fold_config.items():
                print(f"\t>separating fold {n_fold}...")
                fold = labels[studio.isin(studio_num)]  # .drop("Unnamed: 0", axis=1)
                fold.to_csv(f"{self.download_root}/fold{n_fold}.csv", index=False)
            print(">finished separating folds")

    def _get_labels(self, mode="fold"):
        if not os.path.exists(f"{self.download_root}/labels.csv") or \
                not all(os.path.exists(f"{self.download_root}/fold{fold}.csv") for fold in self.fold_config.keys()):
            self._prepare_labels()
        if mode.lower().strip() == "fold":
            return {i: pd.read_csv(f"{self.download_root}/fold{i}.csv") for i in self.fold_config.keys()}
        elif mode.lower().strip() == "full":
            return pd.read_csv(f"{self.download_root}/labels.csv")
        else:
            raise KeyError(f"Invalid mode for getting labels. Expecting `fold|full` but got {mode}")

    def _download(self):
        # download dataset
        if not all(os.path.exists(f"{self.download_root}/{studio}.zip") for studio in self.github_url.keys()):
            print("+-----------------------------------+")
            print("| Downloading dataset...            |")
            print("+-----------------------------------+\n")
        for f, download_url in self.github_url.items():
            if not os.path.exists(f"{self.download_root}/{f}.zip"):
                print(f">downloading {f}.zip ...")
                out_name = os.path.join(self.download_root, f"{f}.zip")
                try:
                    wget.download(url=download_url, out=f"{out_name}", bar=wget.bar_adaptive)
                except HTTPError:
                    raise HTTPError(f"404 Error: Cannot download {download_url}")
            else:
                pass
        # download labels
        if not os.path.exists(f"{self.download_root}/labels.json"):
            try:
                wget.download(url=self.labels_url, out=f"{self.download_root}/labels.json", bar=wget.bar_adaptive)
            except HTTPError:
                raise HTTPError(f"404 Error: Cannot download {self.labels_url}")
        print("Finished Downloading Dataset\n")

        # extract
        stop_check = False
        for studios in self.fold_config.values():
            if stop_check:
                break
            for studio in studios:
                if not os.path.exists(f"{self.download_root}/{studio}"):
                    print("+-----------------------------------+")
                    print("| Extracting dataset...             |")
                    print("+-----------------------------------+\n")
                    stop_check = True
                    break
        for f in sorted(glob(f"{self.download_root}/*.zip")):
            if "studio" in f:
                start, end = os.path.basename(f).split(".")[0].replace("studio", "").split("-")
                directories = [f"{self.download_root}/studio{i:03d}" for i in range(int(start), int(end)+1)]
                if all(os.path.exists(d) for d in directories):
                    continue
            elif "zoom" in f:
                start, end = os.path.basename(f).split(".")[0].replace("zoom", "").split("-")
                directories = [f"{self.download_root}/zoom{i:03d}" for i in range(int(start), int(end) + 1)]
                if all(os.path.exists(d) for d in directories):
                    continue
            elif "labels" in f:
                json_files = [f"{z.replace('.zip', '_label.json')}"
                              for z in sorted(glob(f"{self.download_root}/*.zip")) if "labels" not in z]
                if all(os.path.exists(d) for d in json_files):
                    continue
            else:
                raise NameError(f"Error on name {f}")
            print(f">unzipping {f}...")
            assert os.path.exists(f)
            with zipfile.ZipFile(f) as zip_ref:
                zip_ref.extractall(self.download_root)


# read JSON
def read_json(json_path: str) -> Dict[str, dict]:
    """Read label JSON"""
    with open(json_path, 'r') as f:
        data = json.load(f)
    assert len(set([len(x) for x in data.values()])) == 1, 'Length of the json object values not equal to 1'
    return {k: v[0] for k, v in data.items()}


# labels manipulation
def convert_to_softlabel(evals: List[str]) -> List[float]:
    """Converts a list of emotion into distribution"""
    softlabel = [0 for _ in range(len(emo2idx.keys()))]
    for emo in evals:
        if emo.lower() in ['none', 'other']:
            continue
        softlabel[emo2idx[emo]] += 1
    if np.sum(softlabel) != 0:
        softlabel = np.array(softlabel) / np.sum(softlabel)
    return softlabel


def get_score_from_emo_list(emo_list: List[List[str]]) -> np.ndarray:
    """Aggregate a list of evaluations (which are a list of emotions) into distribution"""
    return np.mean([convert_to_softlabel(evals) for evals in emo_list], axis=0)


def get_agreements(data: Dict[str, dict]) -> Dict[str, np.array]:
    """Get agreement distribution from provided label"""
    softlabel = {k: get_score_from_emo_list(v['annotated']) for k, v in data.items()}
    return softlabel


def convert_to_hardlabel(agreement_dist: List[float], thresh: float = 0.7) -> Union[np.ndarray, int]:
    """convert list"""
    if max(agreement_dist) < thresh:
        return -1
    if pd.Series(agreement_dist).value_counts().sort_index(ascending=False).iloc[0] > 1:
        return -1
    else:
        return np.argmax(agreement_dist)



import argparse
import os

from torch.utils.data import DataLoader
from torchvision.transforms import Compose

from vistec_ser.utils.utils import load_yaml, read_config
from vistec_ser.data.ser_slice_dataset import SERSliceTestDataset
from vistec_ser.data.datasets.emodb import EmoDB
from vistec_ser.data.features.transform import FilterBank
from vistec_ser.evaluation.evaluate import evaluate_slice_model
from vistec_ser.models.network import CNN1DLSTMSlice


def run_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--config-path", type=str, help="path to config file")
    parser.add_argument("--checkpoint-path", type=str, help="path to model checkpoint")
    return parser.parse_args()


def main(args: argparse.Namespace):
    # configure parser
    config_path = args.config_path
    checkpoint_path = args.checkpoint_path
    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Config path not found at {config_path}")
    if not os.path.exists(checkpoint_path):
        raise FileNotFoundError(f"Checkpoint not found at {checkpoint_path}")
    config = load_yaml(config_path)
    hparams, _ = read_config(config)

    # load model
    model = CNN1DLSTMSlice.load_from_checkpoint(
        checkpoint_path=checkpoint_path,
        hparams=hparams)

    # load dataset
    feature_config = config.get("feature", {})
    emodb_config = config.get("emodb", {})
    emodb = EmoDB(**emodb_config, **feature_config)
    emodb.prepare_labels()

    # prepare dataloader
    transform = Compose([FilterBank(
        frame_length=emodb.frame_length,
        frame_shift=emodb.frame_shift,
        num_mel_bins=emodb.num_mel_bins)])
    emodb_loader = DataLoader(SERSliceTestDataset(
        csv_file=emodb.label_path,
        sampling_rate=emodb.sampling_rate,
        max_len=emodb.max_len,
        center_feats=emodb.center_feats,
        scale_feats=emodb.scale_feats,
        transform=transform
    ), batch_size=1)

    # evaluate
    print("\n>>Evaluating Model...\n")
    wa, ua, cm = evaluate_slice_model(model=model, test_dataloader=emodb_loader, n_classes=emodb.n_classes)
    template = f"Confusion Matrix:\n{cm.numpy()}\nWeighted Accuracy: {wa * 100:.2f}%\nUnweighted Accuracy: {ua * 100:.2f}%"
    print(template)
    open(f"{emodb.experiment_dir}/results.txt", "a").write(f"{wa * 100:.2f},{ua * 100:.2f}\n")
    open(f"{emodb.experiment_dir}/confusion_matrix.txt", "a").write(f"{cm.numpy()}")


if __name__ == '__main__':
    main(run_parser())



from typing import Union

import pandas as pd
import torch
import torchaudio
from torch.utils.data import Dataset
from torchaudio.compliance import kaldi

from .features.padding import pad_dup


class SERDataset(Dataset):
    """Speech Emotion Recognition Dataset
    Computing features on the fly with truncated batch
    """

    def __init__(
            self,
            csv_file: Union[str, pd.DataFrame],
            max_len: int,
            sampling_rate: int = 16000,
            emotions=None,
            transform=None):
        """
        Args:
            csv_file (string): Path to the csv file with annotations.
            max_len (int): Maximum length (in second) of audio file. Others will be truncated
            sampling_rate (int): target sampling rate of audio. If this isn't match with file's sampling rate,
                result will be resampled into target sampling rate
            transform (callable, optional): Optional transform to be applied
                on a sample.
        """
        assert isinstance(max_len, int)
        assert isinstance(sampling_rate, int)
        if emotions is None:
            self.emotions = ["neutral", "anger", "happiness", "sadness"]
        else:
            self.emotions = emotions
        if isinstance(csv_file, str):
            self.audios = pd.read_csv(csv_file)
        else:
            self.audios = csv_file
        self.audios = self.audios[self.audios["EMOTION"].str.lower().isin(self.emotions)]
        self.n_classes = len(self.emotions)
        self.sampling_rate = sampling_rate
        self.max_len = max_len
        self.transform = transform

    def __len__(self):
        return len(self.audios)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # df contains 2 columns, [PATH, EMOTION]
        audio_path = self.audios.iloc[idx, 0]
        emotion = self.audios.iloc[idx, 1]
        audio, sample_rate = torchaudio.backend.sox_backend.load(audio_path)

        # initial preprocess
        # convert to mono, resample, truncate
        audio = torch.unsqueeze(audio.mean(dim=0), dim=0)  # convert to mono
        if sample_rate != self.sampling_rate:
            audio = kaldi.resample_waveform(audio, orig_freq=sample_rate, new_freq=self.sampling_rate)

        # emotion to label
        emotion = self.emotions.index(emotion.lower().strip())

        # extract features
        sample = {'feature': audio, 'emotion': emotion}
        if self.transform:
            sample = self.transform(sample)
        feature = sample["feature"]

        # truncated
        sequence_length = self.max_len * 100  # 100 frame = 1 sec
        if feature.shape[-1] > sequence_length:
            feature = feature[:, :sequence_length]
        else:
            feature = pad_dup(feature, sequence_length)
        sample = {"feature": feature, "emotion": emotion}

        return sample



from typing import Dict, List
import os

import torch.nn.functional as F
import torch

from ..models.base_model import BaseSliceModel
from ..models.network import CNN1DLSTMSlice
from ..utils.utils import read_config, load_yaml
from ..data.datasets.thaiser import ThaiSERDataModule


def setup_server(config_path: str):
    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Config file `{config_path}` not found.")

    config = load_yaml(config_path)
    inference_config = config.get("inference", {})

    temp_dir = inference_config.get("temp_dir", "./inference_temp")
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)

    if "checkpoint_path" not in inference_config.keys():
        raise KeyError(f"Error: checkpoint_path not defined")
    checkpoint_path = inference_config["checkpoint_path"]
    if not os.path.exists(checkpoint_path):
        raise FileNotFoundError(f"Checkpoint `{checkpoint_path}` not found.")

    hparams, module_params = read_config(config)
    thaiser_module = ThaiSERDataModule(**module_params)
    model = CNN1DLSTMSlice.load_from_checkpoint(checkpoint_path=checkpoint_path, hparams=hparams)
    model.eval()

    return model, thaiser_module, temp_dir


def infer_sample(model: BaseSliceModel, sample: List[Dict[str, torch.Tensor]], emotions=List[str]):
    name = os.path.basename(sample[0]["emotion"][0])
    final_logits = torch.stack([model(chunk["feature"]) for chunk in sample]).mean(dim=0)
    assert len(final_logits) == 1
    emotion_prob = F.softmax(final_logits[0], dim=-1)
    assert len(emotions) == len(emotion_prob), f"Number of emotion is not equal: len(emotions) = {len(emotions)}, " \
                                               f"len(final_logits) = {len(final_logits)} "
    emotion_prob = {emotion: f"{prob*100:.2f}" for emotion, prob in zip(emotions, emotion_prob)}
    return {"name": name, "prob": emotion_prob}



import torch
import torch.nn as nn
from torch.nn import functional as F


class PyramidLSTM(nn.Module):
    def __init__(
            self,
            input_dim: int,
            hidden_dim: int,
            bidirectional: bool = True,
            dropout: float = 0.):
        super().__init__()
        self.rnn = nn.LSTM(
            input_size=input_dim * 2,
            hidden_size=hidden_dim,
            bidirectional=bidirectional,
            dropout=dropout,
            batch_first=True
        )

    def forward(self, x: torch.Tensor):
        batch_size, time_dim, feat_dim = x.shape
        x = x.contiguous().view(batch_size, int(time_dim / 2), feat_dim * 2)  # stack features
        x, (h, c) = self.rnn(x)
        return x, (h, c)


class AttentionLSTM(nn.Module):
    def __init__(
            self,
            input_dim: int,
            hidden_dim: int,
            output_dim: int,
            bidirectional: bool = True,
            dropout: float = 0.):
        super().__init__()
        self.rnn = nn.LSTM(
            input_size=input_dim,
            hidden_size=hidden_dim,
            bidirectional=bidirectional,
            dropout=dropout,
            batch_first=True
        )
        if bidirectional:
            hidden_dim = hidden_dim * 2
        self.logits = nn.Linear(hidden_dim, output_dim)

    def dot_attention(self, keys: torch.Tensor, query: torch.Tensor):
        """
        alpha = softmax(qK) * V
        where
            - q (query) is LSTM last hidden states
            - K (keys) = V (values) are LSTM sequence
        """
        batch_size = query.shape[1]
        query = query.reshape(batch_size, -1)
        attn_weights = torch.bmm(keys, query.unsqueeze(2)).squeeze(2)
        attn_weights = F.softmax(attn_weights, dim=1)
        output = torch.bmm(keys.transpose(1, 2), attn_weights.unsqueeze(2)).squeeze(2)
        return output

    def forward(self, x):
        x, (h, c) = self.rnn(x)
        x = self.dot_attention(x, h)
        x = self.logits(x)
        return x



import setuptools

requirements = [
    "chardet<4.0,>=2.0",
    "torch>=1.7.1",
    "torchaudio>=0.7.2",
    "torchvision>=0.8.2",
    "pytorch-lightning>=1.2.0",
    "pandas>=1.1.5",
    "numpy>=1.19.2,<1.20.0",
    "soundfile>=0.10.3",
    "PyYAML!=5.4.*,>=5.1",
    "wget>=3.2",
    "fastapi>=0.63.0",
    "aiofiles>=0.6.0",
    "python-multipart>=0.0.5",
    "uvicorn>=0.13.4"
]

setuptools.setup(
    name="vistec-ser",
    version="0.4.3a2",
    author="Chompakorn Chaksangchaichot",
    author_email="chompakorn.cc@gmail.com",
    description="Speech Emotion Recognition models and training using PyTorch",
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/tann9949/vistec-ser',
    packages=setuptools.find_packages(include=['vistec_ser*']),
    install_requires=requirements,
    classifiers=[
        # "2 - Pre-Alpha", "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6'
)



from glob import glob
from typing import List
import os

import wget


class EmoDB(object):
    def __init__(self,
                 sampling_rate: int = 16000,
                 num_mel_bins: int = 40,
                 frame_length: int = 50,
                 frame_shift: int = 10,
                 max_len: int = 3,
                 center_feats: bool = True,
                 scale_feats: bool = True,
                 emotions: List[str] = None,
                 download_dir: str = None,
                 experiment_dir: str = None):
        if emotions is None:
            emotions = ["neutral", "anger", "happiness", "sadness"]
        # config download dir
        if download_dir is None:
            self.download_root = f"{os.path.expanduser('~')}/vistec-ser_tmpfiles/vistec"
        else:
            self.download_root = f"{download_dir}/vistec-ser_tmpfiles/emodb"
        if not os.path.exists(self.download_root):
            os.makedirs(self.download_root)

        # config experiment dir
        if experiment_dir is None:
            self.experiment_root = f"{os.path.expanduser('~')}/vistec-ser_tmpfiles/exp_emodb"
        else:
            self.experiment_root = f"{experiment_dir}"
        self.experiment_dir = f"{self.experiment_root}"
        if not os.path.exists(self.experiment_dir):
            os.makedirs(self.experiment_dir)

        self.download_url = "http://www.emodb.bilderbar.info/download/download.zip"
        self.emotion_mappings = {'N': 'neutral', 'W': 'anger', 'F': 'happiness', 'T': 'sadness'}
        self.label_path = f"{self.download_root}/labels.csv"
        self.test_speaker = ["09", "15"]
        self.val_speaker = ["12", "10"]

        self.max_len = max_len
        self.sampling_rate = sampling_rate
        self.frame_length = frame_length
        self.frame_shift = frame_shift
        self.num_mel_bins = num_mel_bins
        self.center_feats = center_feats
        self.scale_feats = scale_feats
        self.sec_to_frame = 10 * self.frame_shift

        self.emotions = emotions
        self.n_classes = len(self.emotions)

    def download(self):
        # download
        if not os.path.exists(f"{self.download_root}/download.zip"):
            print(">downloading dataset...")
            wget.download(url=self.download_url, out=f"{self.download_root}/download.zip", bar=wget.bar_adaptive)
        # unzip
        if not os.path.exists(f"{self.download_root}/emo-db"):
            print(">unzipping data...")
            os.system(f"unzip -q {self.download_root}/download.zip -d {self.download_root}/emo-db")
        if not os.path.exists(f"{self.label_path}"):
            print(">preparing labels...")
            labels = ["PATH, EMOTION\n"]
            for wav in glob(f"{self.download_root}/emo-db/*/*.wav"):
                key = wav.split('.')[0][-2]
                if key not in self.emotion_mappings.keys():
                    continue
                emotion = self.emotion_mappings[key]
                wav = os.path.abspath(wav)
                labels.append(f"{wav},{emotion}\n")
            open(self.label_path, "w").writelines(labels)

    def prepare_labels(self):
        self.download()
